package codelet;

import RuntimeEnv.RuntimeEnvironment;
import exceptions.*;
import org.jdom2.Element;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.firefox.FirefoxDriver;
import org.yawlfoundation.yawl.elements.data.YParameter;
import org.yawlfoundation.yawl.resourcing.codelets.AbstractCodelet;
import org.yawlfoundation.yawl.resourcing.codelets.CodeletExecutionException;
import rpa.AbstractSeleniumScript;
import rpa.ScriptModel;
import rpa.SeleniumScript;
import util.RPASkriptFinder;
import util.WebDriverFactory;
import util.XMLTransformer;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * This codelet can be used to execute Selenium RPA Scripts, which extends AbstractSeleniumScript
 * @author Simon Rempel
 */

public class SeleniumRPACodelet extends AbstractCodelet {

    //create var for strategy pattern, is used to choose rpa script flexible while runtime
    private SeleniumScript script = null;
    private WebDriver webDriver = null;

    public SeleniumRPACodelet(){
        super();
        setDescription("This codelet can be used to execute Selenium RPA scripts.<br> " +
                "Please keep attention to follow the instructions or use the manual:<br> " +
                "Provided Input: RPAScriptName (string) - name of the RPAScript.<br>" +
                "Custom Input: codeletInput (custom type) - input the RPAScript needs to proceed<br>" +
                "     please create the var if you want do deliver data, see in the RPAScripts manual if you want the wished data type.<br>" +
                "Provided Output: Exception (string) - gets triggered if some exception gets throwed<br>" +
                "     if some exception gets thrown, Exception shows the message, if not Exception is empty.<br>" +
                "Provided Output: ExecutionFailed (boolean) - Shows if the Execution of the RPAScript failed<br>" +
                "     if execution failed, set boolean true and set the messsage in var Exception, else false.<be>" +
                "     This boolean can be used to design exception handling in the yawl sepcification" +
                "Custom Output: codeletOutput (custom type) - outut the RPAScript delivers<br>" +
                "     please create the var if you want do get data, see in the RPAScripts manual if you want the wished data type.<br>" +
                "<br>" +
                "For further details contact your admin.");
    }


    /**
     * Executionmethod: calls needed methods to execute a rpa script
     * works with exception to prevent crash while runtime
     * @param inData The calling workitem's data
     * @param inParams RPA scripts name and codelets input data
     * @param outParams Exception message, execution result ( failed or not) and codelets output data
     * @return the result of the rpa scripts as a jdom element
     * @throws CodeletExecutionException
     */
    @Override
    public Element execute(Element inData, List<YParameter> inParams,
                           List<YParameter> outParams) throws CodeletExecutionException {

        setInputs(inData, inParams, outParams);

        //calls every needed methods, if any exception occurs abort and return boolean as codletOutput.
        //sets Exception to Exceptions Message and ExecutionFailed to true to make YAWL exception handling possible
        try {

            //sets the wished script with strategty pattern
            this.setScript();

            //checks every necessary condition, like delivered and necessary data types
            this.getGuardCheck();

            //this.setWebDriver(inData);

            //executes the rpa script if everything is fine
            this.generelExecution(inData);



        //catches exception if anything occured and delivers default data
        } catch (Exception occuredException){

            //get message to show user in yawl net
            occuredException.printStackTrace();
            System.out.println("Exception Orccured. Message : " + occuredException.getMessage());
            setParameterValue("Exception", occuredException.getMessage());

            //sets execution failed to true to allow exception handling in yawl net
            setParameterValue("ExecutionFailed", "true");

            //set codletOutput if available to default data
            try{
                getOutputParameter("codeletOutput");


                setParameterValue("codeletOutput", getOutputParameter("codeletOutput").getDefaultValue());


                System.out.println(getOutputParameter("codeletOutput").getDefaultValue());
                // getOutputParameter("codeletOutput").setDataTypeAndName("boolean", "codeletOutput", XSD_NAMESPACE );
                //setParameterValue("codeletOutput","false");

            }
            //no codeletOutput was delivered
            catch(Exception noCodeletOutput) {
            }
        }

        //returns output data
        return getOutputData();
    }

    //general execution method, distinguishes between rpa script calls
    private void generelExecution(Element inData) throws RPAExecutionException, NoGeckoDriverFoundException {

        //tries to call right rpa script, if any exception occurs, throw Exception with information
        try{

            //calls rpa script without any input or output data
            if(!script.hasInput() && !script.hasOutput()){
                executeWithoutInputOrOutput(inData, inData.getChildText("GeckodriverPath"));
            }

            //calls rpa script with input but without output data
            if(script.hasInput() && !script.hasOutput()){
                executeWithInputButWithoutOutput(inData, inData.getChildText("GeckodriverPath"));
            }

            //calls rpa script without input but with output data
            if(!script.hasInput() && script.hasOutput()){
                executeWithoutInputButWithOutput(inData, inData.getChildText("GeckodriverPath"));
            }

            //calls rpa script with input and with output data
            if(script.hasInput() && script.hasOutput()){
                executeWithInputAndWithOutput(inData, inData.getChildText("GeckodriverPath"));
            }
            //get message to show user no exception occured
            setParameterValue("Exception","none");

            //get executions sucess for exception handling
            setParameterValue("ExecutionFailed", "false");

            //throws exception if anything went wrong while excecuting one possibility
        }catch (Exception e){
            e.printStackTrace();
            throw new RPAExecutionException("Some Exception occured while executing the RPA script. Please contact your admin. Chosen RPAScript does not work properly");
        }
    }
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    //Execution block
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    private void executeWithoutInputOrOutput(Element inData, String webdriver)throws Exception{
        script.execute(XMLTransformer.XMLtoString(inData), webdriver);
    }

    private void executeWithoutInputButWithOutput(Element inData, String webdriver)throws Exception {

        String result = script.execute(XMLTransformer.XMLtoString(inData), webdriver);
        Element codeletOutput = XMLTransformer.stringToXMLDocument(result).getRootElement();
        //if the output element has no children insert just the value
        if (codeletOutput.getChildren().size() == 0) {
            setParameterValue("codeletOutput", codeletOutput.getValue());
        }
        //if the output element has children insert all of them as value
        else {
            setParameterValue("codeletOutput", XMLTransformer.XMLtoStringWithoutRootElement(codeletOutput));
        }
    }

    private void executeWithInputButWithoutOutput(Element inData, String webdriver)throws Exception {
        script.execute(XMLTransformer.XMLtoString(inData.getChild("codeletInput")), webdriver);
    }

    private void executeWithInputAndWithOutput(Element inData, String webdriver) throws Exception {

        String result = script.execute(XMLTransformer.XMLtoString(inData.getChild("codeletInput")), webdriver);
        Element codeletOutput = XMLTransformer.stringToXMLDocument(result).getRootElement();
        //if the output element has no children insert just the value
        if (codeletOutput.getChildren().size() == 0) {
            setParameterValue("codeletOutput", codeletOutput.getValue());
        }
        //if the output element has children insert all of them as value
        else {
            setParameterValue("codeletOutput", XMLTransformer.XMLtoStringWithoutRootElement(codeletOutput));
        }
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    //Execution block END
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


    public void setWebDriver(Element inData) throws NoGeckoDriverFoundException{
        try {
            webDriver = WebDriverFactory.webDriverFactoryFireFox(inData.getChildText("GeckodriverPath"));
        } catch (Exception e) {
            throw new NoGeckoDriverFoundException("No GeckoDriver Found, please look in control center or contact admin.");
        }
    }

    //set right RPA Script with delivered name
    public void setScript() throws NoSuchRPAScriptException{
        try {
            RPASkriptFinder.getRPAAgentsClass(getValue("RPAScriptName"));
            RuntimeEnvironment.setExecutionMethods();
            script = new SeleniumScript();

        } catch (Exception e) {
            throw  new NoSuchRPAScriptException(e.getMessage());
        }
    }

    private void getGuardCheck() throws LackingInputOrOutputException, NoSuchRPAScriptException {

        if(script==null){
            throw new NoSuchRPAScriptException("There was no Script in the codelets process, please contact your admin.");
        }

        //check if necessary input/output parameters were delivered with right data types
        try {

            if (script.hasInput()) {
                String inputDataType = getInputParameter("codeletInput").getDataTypeName();
                if (!inputDataType.equals(script.getInputDataType())) {
                    throw new LackingInputOrOutputException("Wrong Data Type for codeletInput, needed: " +
                            script.getInputDataType() + ", got :" + inputDataType);
                }
            }

            if (script.hasOutput()) {
                String outputDataType = getOutputParameter("codeletOutput").getDataTypeName();
                if (!outputDataType.equals(script.getOutputDataType())) {
                    throw new LackingInputOrOutputException("Wrong Data Type for codeletOutput, needed: " +
                            script.getOutputDataType() + ", got :" + outputDataType);
                }
            }
        }catch (Exception e){
            e.printStackTrace();
            throw new LackingInputOrOutputException(e.getMessage());
        }

        //check if unnecessary output parameter was delivered
        boolean throwUnnecessaryParameterException = false;
        try{
            getOutputParameter("codeletOutput");
            throwUnnecessaryParameterException = script.hasOutput() ? false : true;
        } catch (Exception e){}

        if(throwUnnecessaryParameterException){
            throw new LackingInputOrOutputException("There was a codeletOutput parameter delivered, but the Script doesent delivers any output.");
        }
    }

    public List<YParameter> getRequiredParams() {

        List<YParameter> params = new ArrayList<YParameter>();

        //set parameter for right rpa script
        YParameter paramRPAName = new YParameter(null, YParameter._INPUT_PARAM_TYPE);
        paramRPAName.setDataTypeAndName("string", "RPAScriptName", XSD_NAMESPACE );
        paramRPAName.setDocumentation("Name of the RPA script which should be executed.");
        params.add(paramRPAName);

        //set parameter for exclet trigger
        YParameter paramExceptionTrigger = new YParameter(null, YParameter._OUTPUT_PARAM_TYPE);
        paramExceptionTrigger.setDataTypeAndName("string", "Exception", XSD_NAMESPACE );
        paramExceptionTrigger.setDocumentation("Trigger name for exclet-rules.");
        params.add(paramExceptionTrigger);

        //set parameter for execution succes
        YParameter paramExecutionFailed = new YParameter(null, YParameter._OUTPUT_PARAM_TYPE);
        paramExecutionFailed.setDataTypeAndName("boolean", "ExecutionFailed", XSD_NAMESPACE );
        paramExecutionFailed.setDocumentation("Shows wheter execution failed or not");
        params.add(paramExecutionFailed);

        YParameter param = new YParameter(null, YParameter._INPUT_PARAM_TYPE);
        param.setDataTypeAndName("string", "GeckodriverPath", XSD_NAMESPACE);
        param.setDocumentation("Hardcoded path to geckodriver, can be empty if given elsewhere");
        params.add(param);

        return params;
    }


}
